/* $Id: icnctrl.c,v 1.5 1995/10/29 21:44:02 fritz Exp fritz $
 *
 * ICN-ISDN driver for Linux. (Control-Utility)
 *
 * Copyright 1994,95 by Fritz Elfert (fritz@wuemaus.franken.de)
 * Copyright 1995 Thinking Objects Software GmbH Wuerzburg
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * $Log: icnctrl.c,v $
 * Revision 1.5  1995/10/29  21:44:02  fritz
 * Changed to support DriverId's.
 * Added support for leased lines.
 *
 * Revision 1.4  1995/04/23  13:43:02  fritz
 * Changed Copyright.
 *
 * Revision 1.3  1995/02/01  11:03:26  fritz
 * Added Messages, when loading bootcode and protocol.
 *
 * Revision 1.2  1995/01/09  07:41:59  fritz
 * Added GPL-Notice
 *
 *
 */

#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>

#include <isdn.h>
#include "icn.h"

char *cmd;
char *ctrldev;
int  arg_ofs;

void usage() {
  fprintf(stderr,"usage: %s [-d driverId] io [mmio port]             (get/set hw-io)\n",cmd);
  fprintf(stderr,"   or: %s [-d driverId] load <bootcode> <protocol> (load firmware)\n",cmd);
  fprintf(stderr,"   or: %s [-d driverId] leased <on|off>            (Switch interface\n",cmd);
  fprintf(stderr,"                                                       into Leased-Line-Mode)\n");
  exit(-1);
}

int main(int argc, char **argv) {
  int fd;
  FILE *code;
  int mmio,port;
  int ac;
  u_char buf[0x10000];
  isdn_ioctl_struct ioctl_s;

  cmd = strrchr(argv[0],'/');
  cmd = (cmd == NULL)?argv[0]:++cmd;
  if (argc>1) {
    if (!strcmp(argv[1],"-d")) {
      strcpy(ioctl_s.drvid,argv[2]);
      arg_ofs = 3;
    } else {
      ioctl_s.drvid[0] = '\0';
      arg_ofs = 1;
    }
  } else
    usage();
  ac = argc - (arg_ofs-1);
  fd = open("/dev/isdnctrl",O_RDWR);
  if (fd < 0) {
    perror(ctrldev);
    exit(-1);
  }
  if (!strcmp(argv[arg_ofs],"leased")) {
    if (ac == 3) {
      ioctl_s.arg = 0;
      if ((!strcmp(argv[arg_ofs+1],"on")) ||
          (!strcmp(argv[arg_ofs+1],"yes"))  )
	ioctl_s.arg = 1;
      if ((ioctl(fd,ICN_IOCTL_LEASEDCFG+ISDN_IOCTL_DRVIOCTL,&ioctl_s))<0) {
	perror("ioctl LEASEDCFG");
	exit(-1);
      }
      close(fd);
      return 0;
    }
  }
  if (!strcmp(argv[arg_ofs],"io")) {
    if (ac == 4) {
      if (sscanf(argv[arg_ofs+1],"%i",&mmio)!=1)
	usage();
      if (sscanf(argv[arg_ofs+2],"%i",&port)!=1)
	usage();
      ioctl_s.arg = mmio;
      if ((mmio = ioctl(fd,ICN_IOCTL_SETMMIO+ISDN_IOCTL_DRVIOCTL,&ioctl_s))<0) {
	perror("ioctl SETMMIO");
	exit(-1);
      }
      ioctl_s.arg = port;
      if ((port = ioctl(fd,ICN_IOCTL_SETPORT+ISDN_IOCTL_DRVIOCTL,&ioctl_s))<0) {
	perror("ioctl SETPORT");
	exit(-1);
      }
    }
    if ((mmio = ioctl(fd,ICN_IOCTL_GETMMIO+ISDN_IOCTL_DRVIOCTL,&ioctl_s))<0) {
      perror("ioctl GETMMIO");
      exit(-1);
    }
    if ((port = ioctl(fd,ICN_IOCTL_GETPORT+ISDN_IOCTL_DRVIOCTL,&ioctl_s))<0) {
      perror("ioctl GETPORT");
      exit(-1);
    }
    printf("Memory-mapped io at 0x%08lx, port 0x%03x\n",
	   (unsigned long)mmio,(unsigned short)port);
    close(fd);
    return 0;
  }
  if (!strcmp(argv[arg_ofs],"load")) {
    if (ac == 4) {
      if (!(code=fopen(argv[arg_ofs+1],"r"))) {
	perror(argv[arg_ofs+1]);
	exit(-1);
      }
      if (fread(buf,4096,1,code)<1) {
	fprintf(stderr,"Read error on %s\n",argv[arg_ofs+1]);
	exit(-1);
      }
      printf("Loading Bootcode %s ...\n",argv[arg_ofs+1]);
      ioctl_s.arg = (ulong)buf;
      if (ioctl(fd,ICN_IOCTL_LOADBOOT+ISDN_IOCTL_DRVIOCTL,&ioctl_s)<0) {
	perror("ioctl LOADBOOT");
	exit(-1);
      }
      fclose(code);
      if (!(code=fopen(argv[arg_ofs+2],"r"))) {
	perror(argv[arg_ofs+2]);
	exit(-1);
      }
      if (fread(buf,65536,1,code)<1) {
	fprintf(stderr,"Read error on %s\n",argv[arg_ofs+2]);
	exit(-1);
      }
      printf("Loading Protocol %s ...\n",argv[arg_ofs+2]);
      ioctl_s.arg = (ulong)buf;
      if (ioctl(fd,ICN_IOCTL_LOADPROTO+ISDN_IOCTL_DRVIOCTL,&ioctl_s)<0) {
	perror("ioctl SETPORT");
	exit(-1);
      }
      fclose(code);
      close(fd);
      return 0;
    }
    usage();
  }
  usage();
  return 0;
}



